/*  Copyright 1995, Richard M. Troth, all rights reserved.  <plaintext> 
 * 
 *	  Name: webcat.c 
 *		General purpose web object reader, akin to 'cat' 
 *	Author: Rick Troth, Houston, Texas 
 *		Rick Troth, Rice University, Information Systems 
 *	  Date: 1994-Jan-14, 15, 1995-Mar-21, Oct-26 
 */ 
 
#include	<fcntl.h> 
#include	<errno.h> 
#include	"webcat.h" 
 
extern  int 	errno;	/* this is redundant in newer systems */ 
	int	webflag; 
static	char	errtx[256]; 
 
/* ------------------------------------------------------------ _WEBHTTP 
 *  URL mode is HTTP.  Retrieve the object that way. 
 *  (port may vary) 
 */ 
int _webhttp(host,port,path)
  char   *host, *port, *path; 
  {
    int 	s; 
    char	string[256]; 
 
    /*  supply the default HTTP port,  if needed  */ 
    if (*port == 0x00) port = "80"; 
 
    /*  open a socket to the HTTP server  */ 
    (void) sprintf(string,"%s:%s",host,port); 
/*  (void) putline(2,string);  */ 
    s = tcpopen(string,0,0); 
    if (s < 0) 
      { 
	(void) strncpy(errtx,host,255); 
	errtx[255] = 0x00; 
	return s; 
      } 
 
    /*  send a "GET" command  */ 
    (void) sprintf(string,"GET %s",path); 
/*  (void) sprintf(string,"POST %s HTTP/1.0",path);  */ 
/*  (void) netline(s,string);  */ 
    (void) tcpputs(s,string); 
 
    /*  hand back the (reading) socket to our caller  */ 
    return s; 
  } 
 
/* ------------------------------------------------------------- _WEBGTP 
 *  URL mode is "gopher".  Open a socket read from the gopher server. 
 *  (port may vary) 
 */ 
int _webgtp(host,port,path) 
  char   *host, *port, *path; 
  { 
    int 	s; 
    char	string[256]; 
 
    /*  supply default port number and "selector string" (path)  */ 
    if (*port == 0x00) port = "70"; 
    if (strlen(path) > 2) path += 2; 
    else path = ""; 
 
    /*  open a socket to the gopher server  */ 
    (void) sprintf(string,"%s:%s",host,port); 
    s = tcpopen(string,0,0); 
    if (s < 0) 
      { 
	(void) strncpy(errtx,host,255); 
	errtx[255] = 0x00; 
	return s; 
      } 
 
    /*  gopher requests are exceedingly simple  */ 
/*  (void) netline(s,path);  */ 
    (void) tcpputs(s,path); 
 
    /*  hand back the (reading) socket to our caller  */ 
    return s; 
  } 
 
/* ------------------------------------------------------------- _WEBFTP 
 *  URL mode is FTP.  FTP uses a two-stream model,  so hand-off 
 *  to the system-supplied  'ftp'  program to handle them. 
 *  (this could reveal compatibility problems; assume a UNIX shell) 
 */ 
int _webftp(host,port,path) 
  char   *host, *port, *path; 
  { 
    int 	fd[2], rc; 
    char	string[256]; 
    char       *user, *pass, temp[256], *cdir, *p; 
 
    /*  attach 'ftp' as a co-process  */ 
    errno = 0; 
    rc = addpiped("ftp -in",fd); 
    if (rc != 0) return -1; 
    if (fd[0] < 0) return fd[0]; 
    if (fd[1] < 0) return fd[1]; 
 
    /*  connect to specified FTP server  */ 
    (void) sprintf(string,"open %s %s",host,port); 
    (void) putline(fd[1],string); 
 
    /*  identify (as "anonymous")  */ 
    user = DEFAULT_FTP_USER; 
/* 
    (void) sprintf(string,"quote user %s",user); 
    (void) putline(fd[1],string); 
 */ 
 
    /*  supply some password  */ 
    (void) sprintf(temp,"%s@",userid()); 
    pass = temp; 
/* 
    (void) sprintf(string,"quote pass %s",pass); 
    (void) putline(fd[1],string); 
 */ 
    (void) sprintf(string,"user %s %s",user,pass); 
    (void) putline(fd[1],string); 
 
    /*  is this binary or plain text?  */ 
    if (webflag & WEBFLAG_BINARY) { 
/*  (void) sprintf(string,"type i");  */ 
    (void) sprintf(string,"binary"); 
    (void) putline(fd[1],string); } 
 
    /*  parse the path [grunt]  */ 
    if (*path == '/') path++; 
    (void) sprintf(temp,"/%s",path); 
    path = temp; cdir = temp; cdir++; 
 
    /*  split basename from rest  */ 
    while (*path != 0x00) path++; 
    while (*path != '/') path--; 
    *path++ = 0x00; 
 
    /*  CD, CD, ...  */ 
    while (*cdir != 0x00) 
      { 
	p = cdir;	/*  not the most efficient place  */ 
	while (*p != '/' && *p != 0x00) p++; 
	if (*p) *p++ = 0x00; 
	(void) sprintf(string,"cd %s",cdir); 
	(void) putline(fd[1],string); 
	cdir = p; 
      } 
 
    /*  retrieve the file desired  */ 
    (void) sprintf(string,"get %s -",path); 
    (void) putline(fd[1],string); 
 
    /*  queue a 'quit' command,  just to be nice  */ 
    (void) putline(fd[1],"quit"); 
 
    /*  hand back the reading end to our caller  */ 
    return fd[0]; 
  } 
 
/* ------------------------------------------------------------- WEBOPEN 
 *  Try to open the "file" using the system open() function. 
 *  If that fails,  try a readlink(),  and,  whether or not the 
 *  file is a sym-link,  treat it as a URL and fetch it from the Web. 
 *  Parse the URL and open according to the specified protocol. 
 */ 
int webopen(p,f,m) 
  char   *p; 
  int     f; 
  int     m; 
  { 
    char	*mode, *host, *port, *path, *q; 
    int 	fd, rc; 
    char	name[WEBCAT_STRING_BUFSIZ], 
		temp[WEBCAT_STRING_BUFSIZ]; 
    int 	sysopenf; 
 
    /*  set sysopenf to UNIX values from webopenf  */ 
    switch (f & (WO_READ|WO_WRITE)) 
      { 
	case WO_READ:	sysopenf = O_RDONLY; 
	case WO_WRITE:	sysopenf = O_WRONLY; 
	case WO_RDWR:	sysopenf = O_RDWR; 
	default:	sysopenf = O_RDONLY; 
      } 
 
    /*  first try the system open() call  */ 
    fd = open(p,sysopenf,m); 
    if (fd >= 0) 
      { 
	if (webflag & WF_VERBOSE) (void) putline(2,"FILE:"); 
	return fd; 
      } 
 
    /*  try following a sym-link to URL  */ 
    rc = readlink(p,name,WEBCAT_STRING_BUFSIZ); 
    if (rc >= 0) 
      { 
	if (webflag & WF_VERBOSE) (void) putline(2,"LINK:"); 
	name[rc] = 0x00; 
      } 
    else (void) strncpy(name,p,WEBCAT_STRING_BUFSIZ); 
    p = name; 
    errno = 0; 
 
    /*  hold onto this ...  */ 
    (void) strncpy(temp,p,WEBCAT_STRING_BUFSIZ); 
 
    /*  start URL processing with empty strings  */ 
    mode = host = port = path = ""; 
 
    /*  parse the URL  */ 
    for (q = p; *q != 0x00 && *q != ':' && *q != '/'; q++); 
    if (*q == ':') { mode = p; *q++ = 0x00; p = q; } 
    if (*q == '/') { 
	if (q[1] == '/') { host = p; 
    for (q += 2; *q != 0x00 && *q != '/'; *p++ = *q++); 
	    *p = 0x00; 
	    p = q; 
	  } 
      } 
    path = p; 
 
    for (q = host; *q != 0x00 && *q != ':'; q++); 
    if (*q == ':') 
      { 
	*q++ = 0x00; 
	port = q; 
      } 
 
    /*  replace empty strings with defaults  */ 
    if (*mode == 0x00) mode = "HTTP"; 
    if (*host == 0x00) host = "localhost"; 
 
    /*  call the supporting protocol routine  */ 
    switch (*mode) 
      { 
	case 'h': case 'H': 
		if (webflag & WF_VERBOSE) (void) putline(2,"HTTP:"); 
		fd = _webhttp(host,port,path); 
/*		_wo_http(p,f,m);		      */ 
		break; 
	case 'g': case 'G': 
		if (webflag & WF_VERBOSE) (void) putline(2,"GOPHER:"); 
		fd = _webgtp(host,port,path); 
/*		_wo_g(p,f,m);			      */ 
		break; 
	case 'f': case 'F': 
		if (webflag & WF_VERBOSE) (void) putline(2,"FTP:"); 
		fd = _webftp(host,port,path); 
/*		_wo_ftp(p,f,m); 		      */ 
		break; 
	default:  /* Protocol not supported */
	    errno = EPROTONOSUPPORT; 
	    return -1; 
      } 
    if (fd < 0) fd = _wo_pxy(temp,f,m); 
    return fd; 
  } 
 
/* ------------------------------------------------------------------ */
int main(argc,argv) 
  int     argc; 
  char   *argv[]; 
  { 
    int 	i, fd0, fd1, rc, wc; 
    char	buffer[WEBCAT_BUFSIZ]; 
    extern  char   *basename(); 
    int 	webopenf, webflags; 
 
    /*  clear global error text string  */ 
    errtx[0] = 0x00; 
 
    /*  set default flag bits  */ 
    webflags = (int) NULL; 
    webflag = WEBFLAG_BINARY;	/*  default to raw image transfer  */ 
    if (strcmp(basename(argv[0]),"webget") == 0) 
      { 
	webflags |= WF_DOWNLOAD; 
	webflag |= WEBFLAG_DOWNLOAD; 
      } 
    if (strcmp(basename(argv[0]),"webman") == 0) 
      { 
	webflags |= WF_MAN; 
      } 
 
    webopenf = WO_READ; 
    /*  process options  */ 
    for (i = 1; i < argc && argv[i][0] == '-' && 
			    argv[i][1] != 0x00; i++) 
      { 
        switch (argv[i][1]) 
          { 
	    case 'a':	webopenf &= ~WO_BINARY; 
			webopenf &= ~WO_HTML; 
			webflag &= ~WEBFLAG_BINARY; 
			webflag &= ~FLAG_HTML; 
			break; 
	    case 'b': 
	    case 'i':	webopenf |= WO_BINARY; 
			webopenf &= ~WO_HTML; 
			webflag |= WEBFLAG_BINARY; 
			webflag &= ~FLAG_HTML;
			break; 
	    case 'h':   webopenf |= WO_HTML; 
			webopenf &= ~WO_BINARY; 
			webflag |= FLAG_HTML; 
			webflag &= ~WEBFLAG_BINARY; 
			break; 
	    case 'w': 
	    case 'p':	webopenf |= WO_WRITE; 
			break; 
	    case 'v':	webflag |= WF_VERBOSE; 
			(void) putline(2,WEBCAT_VERSION); 
			break; 
          } 
      } 
 
    /*  process objects (local files or URLs or links)  */ 
    for (     ; i < argc; i++) 
      { 
	fd0 = webopen(argv[i],webopenf,0); 
	if (fd0 < 0) 
	  { 
	    if (errtx[0] != 0x00) (void) perror(errtx); 
		else		  (void) perror(argv[i]); 
	  } 
	if (fd0 >= 0) 
	  { 
	    if (webflags & WF_DOWNLOAD) 
	      { 
		char *p; 
		p = basename(argv[i]); 
		fd1 = open(p,O_WRONLY|O_TRUNC|O_CREAT,0666); 
		if (fd1 < 0) (void) perror(p); 
	      } 
	    else fd1 = dup(1); 
	  } 
	if (fd0 >= 0 && fd1 >= 0) 
	  { 
 
/*  post: read stdin here  (not implemented)  */ 
	    if (webopenf & WO_WRITE) 
	      { 
		/*  loop reading stdin into fd0  */ 
	      } 
 
	    if (webflag & WEBFLAG_BINARY) while (1) 
	      { 
		rc = read(fd0,buffer,WEBCAT_BUFSIZ); 
		if (rc < 1) break; 
		wc = write(fd1,buffer,rc); 
		if (wc != rc) break; 
	      } 
	    else while (1) 
	      { 
		rc = tcpgets(fd0,buffer,sizeof(buffer)); 
		if (rc < 0) break; 
		rc = putline(fd1,buffer); 
		if (rc < 0) break; 
	      } 
	  } 
	if (fd0 >= 0) (void) close(fd0); 
	if (fd1 >= 0) (void) close(fd1); 
      } 
 
    (void) close(0); 
    (void) close(1); 
    (void) close(2); 
 
    return 0; 
  } 
 

